@@ -1,11 +1,13 @@
 #ifndef DCGP_SYMBOLIC_REGRESSION_H
 #define DCGP_SYMBOLIC_REGRESSION_H
+#include <algorithm>
 #include <audi/gdual.hpp>
 #include <boost/numeric/conversion/cast.hpp>
 #include <boost/range/algorithm/transform.hpp>
 #include <numeric> // std::accumulate
 #include <pagmo/io.hpp>
 #include <pagmo/population.hpp>
+#include <pagmo/problem.hpp>
 #include <pagmo/types.hpp>
 #define and &&
 #define or ||
@@ -19,21 +21,24 @@
 #include <dcgp/expression.hpp>
 #include <dcgp/kernel_set.hpp>
 #include <dcgp/rng.hpp>
+#include <dcgp/s11n.hpp>
 
 namespace dcgp
 {
 /// A Symbolic Regression problem
 /**
  *
+ * \image html symbolic_regression.jpg "Math Formulae"
  *
- * Symbolic regression is a type of regression analysis that searches the space of mathematical expressions to 
- * find the model that best fits a given dataset, both in terms of accuracy and simplicity 
+ * Symbolic regression is a type of regression analysis that searches the space of mathematical expressions to
+ * find the model that best fits a given dataset, both in terms of accuracy and simplicity
  * (ref: https://en.wikipedia.org/wiki/Symbolic_regression). It also is one of the core applications
  * for Differentiable Cartesian Genetic Programming.
  *
  * This class provides an easy way to instantiate symbolic regression problems as optimization problems having
  * a continuous part (i.e. the value of the parameters in the model) and an integer part (i.e. the representation of
- * the model computational graph). The instantiated object can be used as UDP (User Defined Problem) in the pagmo optimization suite.
+ * the model computational graph). The instantiated object can be used as UDP (User Defined Problem) in the pagmo
+ * optimization suite.
  *
  * The symbolic regression problem can be instantiated both as a single and a two-objectives problem. In the second
  * case, aside the Mean Squared Error, the formula complexity will be considered as an objective.
@@ -182,7 +187,9 @@ public:
                 pagmo::stream(ss, prettier);
                 auto string = ss.str();
                 // We remove whitespaces too
-                l_prettier += static_cast<double>(string.length() - static_cast<decltype(string.length())>(std::count(string.begin(), string.end(), ' ')));
+                l_prettier += static_cast<double>(
+                    string.length()
+                    - static_cast<decltype(string.length())>(std::count(string.begin(), string.end(), ' ')));
             }
             // Here we define the formula complexity
             retval[1] = std::min(l_pretty, l_prettier);
@@ -460,14 +467,56 @@ public:
         return m_cgp;
     }
 
+    /// Model predictions
+    /**
+     * Uses the model encoded in *x* to predict the label of *point*.
+     *
+     * @param[in] point point to be predicted.
+     * @param[in] x chromosome encoding the model.
+     *
+     * @return the predicted label for *point*.
+     */
+    std::vector<double> predict(const std::vector<double> &point, pagmo::vector_double x) const
+    {
+        // Here we set the CGP member from the chromosome
+        set_cgp(x);
+        return m_cgp(point);
+    }
+
+    /// Model predictions
+    /**
+     * Uses the model encoded in *x* to predict the labels of *points*.
+     *
+     * @param[in] points points to be predicted.
+     * @param[in] x chromosome encoding the model.
+     *
+     * @return the predicted labels for *points*.
+     */
+    std::vector<std::vector<double>> predict(const std::vector<std::vector<double>> &points,
+                                             pagmo::vector_double x) const
+    {
+        // This will hold the return value
+        std::vector<std::vector<double>> retval;
+        // Here we set the CGP member from the chromosome
+        set_cgp(x);
+        // We loop over the input points
+        for (decltype(points.size()) i = 0u; i < points.size(); ++i) {
+            retval.push_back(m_cgp(points[i]));
+        }
+        return retval;
+    }
+
     /// Thread safety for this udp
     /**
      * This is set to none as pitonic kernels could be in the inner expression
      */
-    //pagmo::thread_safety get_thread_safety() const
-   // {
-    //    return pagmo::thread_safety::none;
-   // }
+    pagmo::thread_safety get_thread_safety() const
+    {
+        return (*std::min_element(
+                    m_f.begin(), m_f.end(),
+                    [](const auto &a, const auto &b) { return a.get_thread_safety() < b.get_thread_safety(); }))
+            .get_thread_safety();
+    }
 
 private:
     // This setter can be marked const as m_cgp is mutable
@@ -516,6 +565,30 @@ private:
         if (m_f.size() == 0) throw std::invalid_argument("Number of basis functions is 0");
     }
 
+public:
+    template <typename Archive>
+    void serialize(Archive &ar, unsigned)
+    {
+        ar &m_points;
+        ar &m_labels;
+        ar &m_dpoints;
+        ar &m_dlabels;
+        ar &m_deph_symb;
+        ar &m_symbols;
+        ar &m_r;
+        ar &m_c;
+        ar &m_l;
+        ar &m_arity;
+        ar &m_f;
+        ar &m_n_eph;
+        ar &m_multi_objective;
+        ar &m_parallel_batches;
+        ar &m_cgp;
+        ar &m_dcgp;
+        ar &m_cache_fitness;
+    }
+
+private:
     std::vector<std::vector<double>> m_points;
     std::vector<std::vector<double>> m_labels;
     std::vector<std::vector<audi::gdual_d>> m_dpoints;
@@ -539,6 +612,9 @@ private:
     mutable expression<audi::gdual_d> m_dcgp;
     mutable std::pair<pagmo::vector_double, pagmo::vector_double> m_cache_fitness;
     mutable std::pair<pagmo::vector_double, pagmo::vector_double> m_cache_gradient;
-}; // namespace dcgp
+};
 } // namespace dcgp
+
+PAGMO_S11N_PROBLEM_EXPORT_KEY(dcgp::symbolic_regression)
+
 #endif
